class Lookup(object):
    def __init__(self, data: dict):
        self._data = data
    
    def __getattr__(self, key: str):
        value = self._data.get(key)
        if isinstance(value, dict):
            return Lookup(value)
        return value


class Steps(object):
    def __init__(self):
        self._first = None
        self._values = []
        self._misses = []

    def EQUALS(self, datum: any, key: str):
        # necessary as sqlite uses 0 a falsy value for false
        if(datum == 0 or datum):
            if(self.dirty):
                sqls = f'AND {key} = ?'
            else:
                sqls = f'WHERE {key} = ?'
            self._values.append(datum)
            return sqls
        return ''

    def LIKE(self, datum: any, key: str):
        if(datum):
            if(self.dirty):
                sqls = f'AND {key} LIKE ?'
            else:
                sqls = f'WHERE {key} LIKE ?'
            self._values.append(f'%{datum}%')
            return sqls
        return ''

    @property
    def dirty(self) -> bool:
        return bool(self._values)

    @property
    def values(self) -> list:
        return self._values



from os import environ
from sqlite3 import Connection
from heaven import Application, Response


async def cors(req, res: Response, ctx):
    res.headers = 'Access-Control-Allow-Credentials', 'true'


def upsudo(app: Application) -> str:
    db: Connection = app.peek('db')
    username = environ.get('ADMINISTRATOR_USERNAME')
    password = environ.get('ADMINISTRATOR_PASSWORD')

    # credentials table dropped every time so this is possible, what of microservice credentials created after app start?
    try: db.execute(f'INSERT INTO credentials VALUES(?, ?);', (username, password))
    except: pass



from http import HTTPStatus
from sqlite3 import Connection

from fastjsonschema import JsonSchemaException, compile
from heaven import Context, Request, Response
from orjson import dumps, loads

from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import expects
from amebo.models.events import Events
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
def tabulate(req: Request, res: Response, _: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)

    params = ['id', 'action', 'deduper', 'payload', 'timeline']
    _id, _action,  _deduper, _payload, _timeline = [req.params.get(p) for p in params]

    steps = Steps()
    sqls = f'''SELECT
            event, action, payload, deduper, timestamped, COUNT(*) AS results
        FROM events
            {steps.EQUALS(_id, 'event')}
            {steps.LIKE(_action, 'action')}
            {steps.EQUALS(_deduper, 'deduper')}
            {steps.LIKE(_payload, 'payload')}
            {get_timeline(_timeline, steps)}
        GROUP BY
            event, action, payload, deduper, timestamped
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try:
        cursor = db.cursor()
        rows = cursor.execute(sqls, steps.values).fetchall()
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return
    finally:
        cursor.close()

    res.status = HTTPStatus.OK
    res.body = [{
        'event': event,
        'action': action,
        'payload': loads(payload),
        'deduper': deduper,
        'timestamped': timestamped
    } for event, action, payload, deduper, timestamped, results in rows]


@jsonify
@expects(Events)
def insert(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    event: Events = ctx.events

    table = 'events'
    fields = ('action', 'payload', 'deduper', 'timestamped',)

    try:
        cursor = db.cursor()
        row = cursor.execute('''
            SELECT schemata FROM actions WHERE action = ?
        ''', (event.action,)).fetchone()

        if not row:
            res.status = HTTPStatus.UNPROCESSABLE_ENTITY
            res.body = {'error': 'Action can not be used to process any events'}
            return

        schemata = loads(row[0])
        _schemas = req.app.peek('schematas')
        if not _schemas.get(event.action): _schemas[event.action] = compile(loads(schemata))
        validation = _schemas.get(event.action)
        validation(event.payload)
        values = (event.action,
            dumps(event.payload), event.deduper, event.timestamped)

        eventid = cursor.execute(f'''
            INSERT INTO
                {table} ({', '.join(fields)})
            VALUES(?, ?, ?, ?) RETURNING rowid;
        ''', values).fetchone()

        cursor.execute(f'''
            INSERT INTO
                gists(event, subscriber, completed, timestamped)
            SELECT
                {eventid[0]}, subscriber, 0, '{event.timestamped.isoformat()}'
            FROM subscribers WHERE action = ?
        ''', (event.action,));

        db.commit()
    except JsonSchemaException:
        return res.out(HTTPStatus.NOT_ACCEPTABLE, {'error': f'Event payload does not conform to {event.action} schema'})
    except Exception as exc:
        res.status = HTTPStatus.UPGRADE_REQUIRED
        return res.out(HTTPStatus.UPGRADE_REQUIRED, {'error': f'{exc}'})
    finally:
        cursor.close()

    res.status = HTTPStatus.CREATED
    res.body = {
        'event': eventid[0],
        'action': event.action,
        'payload': event.payload,
        'deduper': event.deduper,
        'timestamped': event.timestamped
    }



"""
TODO: change to supersql and group sqlite&pg -> sql so the final option is sql | redis
"""

from constants import PG, REDIS, SQLITE


def register_actions_redis():
    pass


def register_actions_sql():
    pass


def register_applications_redis():...


def register_applications_sql():...


def register_events_redis():
    pass


def register_events_sql():
    pass


def register_subscribers_redis():...


def register_subscribers_sql():...


register_actions = {
    REDIS: register_actions_redis,
    SQLITE: register_actions_sql
}


register_events = {
    REDIS: register_events_redis,
    SQLITE: register_events_sql
}


register_applications = {
    REDIS: register_applications_redis,
    SQLITE: register_applications_sql
}


register_subscribers = {
    REDIS: register_subscribers_redis,
    SQLITE: register_subscribers_sql
}



from asyncio import gather, sleep
from http import HTTPStatus
from sqlite3 import Connection, Cursor

# installed libs
from heaven import Router
from httpx import AsyncClient
from orjson import loads

# src code
from amebo.constants.literals import DB


async def aproko(router: Router):
    db: Connection = router.peek(DB)
    accepters = []
    rejecters = []
    async def notify(endpoint: str, data: dict, passkey: str, gist_id: int):
        headers = {
            'Content-Type': 'application/json',
            'X-Secret-Key': passkey
        }

        try:
            client = AsyncClient()
            result = await client.post(endpoint, json=data, headers=headers)
            if result.status_code not in [HTTPStatus.ACCEPTED, HTTPStatus.OK]:
                rejecters.append(int(gist_id))
            else:
                accepters.append(int(gist_id))
        except Exception as exc:
            rejecters.append(gist_id)
        finally:
            await client.aclose()

        try:
            cursor: Cursor = db.cursor()
            cursor.execute(f'''
                UPDATE
                    gists
                SET
                    completed = 1
                WHERE rowid IN {format(tuple(accepters)).replace( ',)', ')' )};
            ''')
            db.commit()
        except Exception as exc:
            print('Could not update in notify: ', exc)
        finally:
            cursor.close()

    async def traverse():
        cursor = None
        try:
            cursor: Cursor = db.cursor()
            gists = cursor.execute(f'''
                SELECT
                    m.location || s.endpoint AS endpoint, e.payload, m.passkey, g.rowid as gid
                FROM gists AS g JOIN events e ON
                    g.event = e.event
                JOIN subscribers s ON
                    s.subscriber = g.subscriber
                JOIN actions a ON
                    e.action = a.action
                JOIN microservices m ON
                    s.microservice = m.microservice
                WHERE g.completed <> 1
                ORDER BY g.event LIMIT {router.CONFIG('envelope_size')};
            ''').fetchall()

            if len(gists) < router.CONFIG('rest_when'): await sleep(router.CONFIG('idles'))
            await gather(*[notify(endpoint, loads(payload), passkey, gid) for endpoint, payload, passkey, gid in gists])

            db.commit()
        except Exception as exc:
            print('Exception occured: ', exc)
        finally:
            if cursor: cursor.close()

    await traverse()



from datetime import datetime
from json import loads
from typing import Optional, Union

from pydantic import AnyHttpUrl
from pydantic import BaseModel as Model
from pydantic import Field, field_validator

from amebo.models.actions import Action


class Subscriber(Model):
    subscriber: Optional[int]
    action: str
    microservice: str
    endpoint: str
    description: str
    timestamped: datetime = Field(default_factory=datetime.now)


class Gistings(Model):
    attempts: int = 0
    action: Action
    subscriber: Subscriber
    status: int



from functools import wraps
from http import HTTPStatus
from inspect import iscoroutinefunction

from fastjsonschema import compile
from heaven import Context, Request, Response
from orjson import dumps, loads
from pydantic import BaseModel

from amebo.utils.structs import Lookup


def expects(Model: BaseModel):
    def wrapper(func):
        async def delegate(req: Request, res: Response, ctx: Context):
            appjson, ctype = 'application/json', 'content-type'
            res.headers = ctype, appjson
            if(req.headers.get(ctype) != appjson):
                res.status = HTTPStatus.IM_A_TEAPOT
                res.body = {'error': 'your request does not know json kung-fu'}
                return

            try:
                body = loads(req.body)
                model = Model(**body)
                mname = model.__class__.__name__.lower()
                ctx.keep(mname, model)
            except Exception as exc:
                error = exc.errors()[0]
                res.status = HTTPStatus.BAD_REQUEST
                print('-' * 20, error)
                res.body = {'error': f"{error.get('loc')[0]} - {error.get('msg')}"}
                print('-' * 10, res.body)
                return

            if iscoroutinefunction(func): await func(req, res, ctx)
            else: func(req, res, ctx)
        return delegate
    return wrapper


def inspects(schema: dict):
    validation = compile(schema)  # validictory?
    def decorator(func):
        @wraps(func)
        async def delegate(req: Request, res: Response, ctx: Context):
            try: payload = validation(loads(req.body))
            except: return res.out(HTTPStatus.NOT_ACCEPTABLE, 'Invalid data detected')
            ctx.keep('data', Lookup({key: value for key, value in payload.items() if value}))
            return await func(req, res, ctx)
        return delegate
    return decorator



from datetime import datetime, timedelta

from jwt import decode, encode
from heaven import Request

from amebo.constants.literals import DEFAULT_PAGINATION


def get_pagination(req: Request):
    try: page = int(req.params.get('page'))
    except: page = 1
    else: page = 1 if page < 0 else page

    try: pagination = int(req.params.get('pagination'))
    except: pagination = DEFAULT_PAGINATION
    else: pagination = DEFAULT_PAGINATION if pagination < 0 else pagination
    return page, pagination


def get_params(params: list, req: Request):
    return [req.params.get(p) for p in params]


def get_timeline(timeline, step_or_filter, column: str = None):
    if timeline:
        dateline = datetime.now()
        value = timeline.lower()
        if value == 'month':
            dateline = dateline - timedelta(days=31)
        elif value == 'week':
            dateline = dateline - timedelta(days=7)
        elif value == 'today':
            dateline = dateline - timedelta(hours=24)

        adjunction = 'AND' if step_or_filter.dirty else 'WHERE'
        return f"{adjunction} {column or 'timestamped'} > DATETIME('{dateline.isoformat()}')"
    return ''


def tokenize(data, sk):
    return encode(data, sk)


def untokenize(token, sk):
    pass



from functools import wraps
from inspect import iscoroutinefunction

from fastjsonschema import compile
from heaven import Context, Request, Response
from orjson import dumps

from amebo.utils.structs import Lookup


def jsonify(func):
    @wraps(func)
    async def delegate(req, res, ctx: Context):
        if(iscoroutinefunction(func)): await func(req, res, ctx)
        else: func(req, res, ctx)
        res.headers = 'Content-Type', 'application/json'
        res.body = dumps(res.body)
    return delegate


def parameters(params: dict):
    def wrapper(func):
        @wraps(func)
        async def proxy(req: Request, res: Response, ctx: Context):
            lookup = {}
            for key, value in params.items():
                if key in req.params: lookup[key] = value(req.params.get(key))
                ctx.keep('params', Lookup(lookup))
            return await func(req, res, ctx)
        return proxy
    return wrapper


# ie. {name: {kind: str, default: ''}}
def queries(schema: dict):
    validator = compile(schema)
    def wrapper(func):
        @wraps(func)
        async def delegate(req: Request, res: Response, ctx: Context):
            qp = {}
            for param in params:
                # dict.values might give you results out of order if key order changed
                kind, default = param.get('kind'), param.get('default')
                try: value = kind(req.queries.get(param))
                except: value = default
                else: qp[param] = value
            ctx.keep('queryparams', qp)
            return await func(req, res, ctx)
        return delegate
    return wrapper



from http import HTTPStatus
from sqlite3 import Connection, Cursor, IntegrityError

# installed libs
from heaven import Context, Request, Response

# src code
from amebo.constants.literals import DB, SECRET_KEY, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import expects
from amebo.utils.helpers import get_pagination, get_timeline, tokenize

from amebo.models.microservices import Credential, Location, Microservice
from amebo.utils.structs import Steps


@jsonify
@expects(Credential)
def authenticate(req: Request, res: Response, ctx: Context):
    def unauthorized():
        res.status = HTTPStatus.UNAUTHORIZED
        res.body = {'error': 'could not authenticate microservice'}

    db: Connection = req.app.peek(DB)
    credential: Credential = ctx.credential

    table = 'credentials'
    username_field = 'username'
    password_field = 'password'
    if credential.scheme == 'token':
        table = 'microservices'
        username_field = 'microservice'
        password_field = 'passkey'

    try:
        cursor: Cursor = db.cursor()
        row = cursor.execute(f'''
            SELECT {username_field}, {password_field} FROM {table}
                WHERE {username_field} = ?
        ''', (credential.username,))
    except Exception as exc:
        return unauthorized()
    else: data = row.fetchone()
    finally: cursor.close()

    if not data: return unauthorized()

    username, password = data
    if password != credential.password: return unauthorized()
    # no feedback if provided if secret key mismatches i.e. continue indicates just that

    token = tokenize({
        'scheme': credential.scheme,
        'username': username,
    }, req.app.CONFIG(SECRET_KEY))
    res.headers = 'Set-Cookie', f'Authorization={token}; Path=/; HttpOnly; Max-Age={60*10}; SameSite=Strict; Secure'
    res.status = HTTPStatus.ACCEPTED
    res.body = None


@jsonify
def tabulate(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    params = ['microservice', 'location', 'timeline']
    _microservice, _location, _timeline = [req.params.get(p) for p in params]
    steps = Steps()

    m = 'microservice'

    sqls = f'''SELECT rowid AS id, microservice, location, timestamped
        FROM microservices
            {steps.LIKE(_microservice, m)}
            {steps.LIKE(_location, 'location')}
            {get_timeline(_timeline, steps)}
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try:
        cursor = db.cursor()
        rows = cursor.execute(sqls, steps.values).fetchall()
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return
    finally:
        if cursor: cursor.close()

    res.status = HTTPStatus.OK
    res.body = [{
        'id': id,
        'microservice': microservice,
        'location': location,
        'passkey': '****************',
        'timestamped': timestamped
    } for id, microservice, location, timestamped in rows]


@jsonify
@expects(Microservice)
def insert(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    microservice: Microservice = ctx.microservice
    try:
        cursor = db.cursor()
        values = (microservice.microservice, str(microservice.address), microservice.passkey, microservice.timestamped)
        cursor.execute('''
            INSERT into microservices(microservice, location, passkey, timestamped) VALUES (?, ?, ?, ?)
        ''', values)
        db.commit()
    except IntegrityError as exc:
        res.status = HTTPStatus.CONFLICT
        res.body = {'error': f'{exc}'}
        return
    except Exception as exc:
        res.status = HTTPStatus.NOT_ACCEPTABLE
        res.body = {'error': f'{exc}'}
        return
    finally:
        cursor.close()
    
    res.status = HTTPStatus.CREATED
    res.body = {
        'name': microservice.microservice,
        'location': str(microservice.address),
        'passkey': microservice.password,
        'timestamped': microservice.timestamped
    }


@jsonify
@expects(Location)
def update(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    microservice = req.params.get('id')
    location: Location = ctx.location

    try:
        cursor = db.cursor()
        cursor.execute(f'''
            UPDATE microservices SET location = ?
                WHERE microservice = ? AND passkey = ?
        ''', (location.location, microservice, location.passkey))
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': 'could not update microservice'}
        return
    finally:
        cursor.close()

    # no feedback if provided if secret key mismatches i.e. continue indicates just that
    res.status = HTTPStatus.ACCEPTED
    res.body = None



from heaven import Context, Request, Response


async def pages(req: Request, res: Response, ctx: Context):
    page = req.params.get('page')
    if page not in ['actions', 'events', 'subscribers', 'microservices', 'gists']:
        page = '404'
    return await res.render(f'{page}.html', req=req)


async def login(req: Request, res: Response, ctx: Context):
    return await res.render('login.html', req=req)



from pathlib import Path
from subprocess import call
from sys import argv


def execute():
    options = argv[1:]
    cwd = Path(__file__).parent.resolve()
    call(['sh', f'{cwd}/amebo.sh', *options])


if __name__ == '__main__':
    execute()



from http import HTTPStatus
from inspect import iscoroutinefunction

from heaven import Context, Request, Response


def authenticated(func):
    async def delegate(req: Request, res: Response, ctx: Context):
        cookie = req.cookies.get('Authorization')
        if not cookie:
            print(req.headers)
            print(req.cookies)
            res.status = HTTPStatus.TEMPORARY_REDIRECT
            res.headers = 'Location', '/'
            return

        if iscoroutinefunction(func): return await func(req, res, ctx)
        else: return func(req, res, ctx)
    return delegate



from os import environ
from sqlite3 import Connection

from heaven import Router


def upsudo(router: Router) -> str:
    db: Connection = router.peek('db')
    username = environ.get('ADMINISTRATOR_USERNAME')
    password = environ.get('ADMINISTRATOR_PASSWORD')
    try:
        # credentials table dropped every time so this is possible
        db.execute(f'''
            INSERT INTO credentials VALUES(?, ?);
        ''', (username, password))
    except: pass



from datetime import datetime
from json import loads
from typing import Optional, Union

from pydantic import AnyHttpUrl
from pydantic import BaseModel as Model
from pydantic import Field, field_validator


class Events(Model):
    action: str = Field(...)
    event: Optional[int] = None
    deduper: str
    payload: Union[str, dict]
    timestamped: datetime = Field(default_factory=datetime.now)

    @field_validator('payload')
    def validate_payload(cls, value: Union[str, dict]):
        if isinstance(value, dict): return value
        return loads(value)
    
    @field_validator('action')
    def _action(cls, value: str):
        if not value: raise ValueError('Action is a required field')
        return value



from os import environ
from sqlite3 import Connection

from heaven import Application
from redis import Redis
from asyncpg import connect as pgconnect

from amebo.constants.literals import DB
from amebo.constants.scripts import initdbscript


ENGINES = {
    'sqlite': lambda *args: Connection(*args),
    'redis': lambda *args: Redis(*args),
    'postgres': lambda *args: pgconnect(*args)
}


async def connect(app: Application):
    engine = app.CONFIG(DB).lower()
    app.keep('engine', engine)
    connection = ENGINES.get(engine)
    db = connection(environ.get('AMEBO_DSN') or 'amebo.db')
    app.keep(DB, db)


async def disconnect(app: Application):
    try: db: Connection = app.peek(DB)
    except Exception as exc: print('Exception in closing db on shutdown: ', exc)
    else: db: db.close()


async def initialize(app: Application):
    """todo: enable switching db backend between redis, pg, sqlite"""
    try:
        db: Connection = app.peek(DB)
        db.executescript(initdbscript)
    except Exception as exc:
        print('Exception in initdb hook: ', exc)


def cache(app: Application):
    app.keep('schematas', {})



from asyncio import (
    get_running_loop,
    sleep,
)
from http import HTTPStatus
from sqlite3 import Connection

# installed libs
from httpx import AsyncClient


async def amebo(db: Connection):
    """This is where washika handles its business"""
    try:
        cursor = db.cursor()
        rows = cursor.execute('SELECT position, action, payload, timestamped FROM events ORDER BY timestamped')
        row = rows.fetchone()
        if row:
            _, action, payload, _ = row
            for result in rows:
                subscribers = cursor.execute(f'SELECT application, endpoint, passkey FROM subscribers WHERE action = {result.action}')
                for subscriber in subscribers:
                    application, endpoint, passkey = subscriber
                    async with AsyncClient() as ahttp:
                        res = await ahttp.post(endpoint, data=payload, headers={'Authorization': f'bearer {passkey}'})
                    if res.status_code != HTTPStatus.ACCEPTED:
                        cursor.execute(f'INSERT INTO subscribers')
    except Exception as exc:
        print(exc)
        cursor.close()


async def amebo_infinite_recursion_loop():
    loop = get_running_loop()
    # await amebo()
    # print('+')
    await sleep(2)
    loop.create_task(amebo_infinite_recursion_loop())



initdbscript = '''
BEGIN;
    DROP TABLE IF EXISTS credentials;

    CREATE TABLE IF NOT EXISTS credentials(
        username varchar(128) primary key,
        password varchar not null
    );

    CREATE TABLE IF NOT EXISTS microservices (
        microservice varchar(128) primary key,
        location varchar(512) not null,
        passkey text not null,
        timestamped varchar(32) not null
    );

    CREATE TABLE IF NOT EXISTS actions (
        action varchar(128) primary key not null,
        microservice varchar(128) NOT NULL REFERENCES microservices(microservice),
        schemata text not null,
        timestamped varchar(32) not null
    );

    CREATE TABLE IF NOT EXISTS events (
        event integer primary key not null,
        action varchar(128) not null references actions(action),
        deduper varchar(128) not null,
        payload text not null,
        timestamped varchar(32) not null,

        UNIQUE(deduper, payload)
    );

    CREATE TABLE IF NOT EXISTS subscribers (
        subscriber integer primary key not null,
        microservice varchar(128) not null references microservices(microservice),
        action varchar(128) not null references actions(action),
        endpoint varchar(256) not null,
        description text not null,
        timestamped varchar(32) not null,

        UNIQUE(microservice, action, endpoint)
    );

    CREATE TABLE IF NOT EXISTS gists (
        event integer references events(event),
        subscriber integer references subscribers(subscriber),
        completed integer not null,
        timestamped varchar(32) not null,

        UNIQUE(event, subscriber)
    );
COMMIT;
'''

insertscript = '''
    INSERT INTO credentials VALUES(?, ?)
'''

tidydatabase = '''
    CREATE TRIGGER IF NOT EXISTS
        prune_gists
    AFTER INSERT ON
        gists
    WHEN
        count(*) > 1_000_000
    BEGIN
        DELETE FROM gists WHERE timestamped < '14 days ago'
    END
'''



from http import HTTPStatus
from sqlite3 import Connection, Cursor

# installed libs
from heaven import Context, Request, Response
from orjson import dumps, loads

from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import expects
from amebo.models.actions import Action
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
def tabulate(req, res, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    params = ['id', 'action', 'microservice', 'schemata', 'timeline']
    _id, _action, _microservice, _schemata, _timeline = [req.params.get(p) for p in params]

    steps = Steps()
    sqls = f'''
        SELECT
            rowid, action, microservice, schemata, timestamped
        FROM
            actions
            {steps.EQUALS(_id, 'rowid')}
            {steps.LIKE(_action, 'action')}
            {steps.LIKE(_microservice, 'microservice')}
            {steps.LIKE(_schemata, 'schemata')}
            {get_timeline(_timeline, steps)}
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try:
        cursor = db.cursor()
        rows = cursor.execute(sqls, steps.values).fetchall()
    except Exception as exc:
        res.status = HTTPStatus.NO_CONTENT
        res.body = None
        return
    finally:
        cursor.close()

    res.status = HTTPStatus.OK
    res.body = [{
        'id': id,
        'action': action,
        'microservice': microservice,
        'schemata': loads(schemata),
        'timestamped': timestamped
    } for id, action, microservice, schemata, timestamped in rows]


@jsonify
@expects(Action)
def insert(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    action: Action = ctx.action

    table = 'actions'
    fields = ('action', 'microservice', 'schemata', 'timestamped',)
    values = (action.action, action.microservice, dumps(action.schemata), action.timestamped)

    try:
        cursor: Cursor = db.cursor()
        microservice = cursor.execute(f'''
            select * from microservices where rowid = ?
        ''', (action.microservice,)).fetchone()
        if not microservice:
            raise ValueError(f'microservice with id {action.microservice} not found')

        cursor.execute(f'''
            INSERT INTO {table}({', '.join(fields)}) VALUES(?, ?, ?, ?)
        ''', values)
        db.commit()
    except Exception as exc:
        res.status = HTTPStatus.UPGRADE_REQUIRED
        res.body = {'error': f'{exc}'}
        return
    finally:
        cursor.close()

    res.status = HTTPStatus.CREATED
    res.body = action.dict()



from asyncpg import create_pool as pgpool
from aiosqlite import connect as liteconnect


class PostgresEngine(object):
    def __init__(self, *args):
        self.__pool = pgpool(*args)

    def create_microservice(self, microservice: str, endpoint: str, passkey: str):
        self.__redis.set(microservice, f'{endpoint} {passkey}')

    def delete_microservice(self, microservice: str):
        self.__redis.delete(microservice)

    def get_microservice(self, microservice: str):
        return self.__redis.get(microservice)

    def save_action(self, action: str, data: dict):
        self.__redis.set(action, data)

    def get_action(self, action: str):
        return self.__redis.get(action)


class LiteEngine(object):
    def __init__(self, *args):
        self.__conn = liteconnect(*args)

    def create_microservice(self, microservice: str, endpoint: str, passkey: str):
        self.__redis.set(microservice, f'{endpoint} {passkey}')

    def delete_microservice(self, microservice: str):
        self.__redis.delete(microservice)

    def get_microservice(self, microservice: str):
        return self.__redis.get(microservice)

    def save_action(self, action: str, data: dict):
        self.__redis.set(action, data)

    def get_action(self, action: str):
        return self.__redis.get(action)



from http import HTTPStatus
from sqlite3 import Connection, Cursor, IntegrityError

from heaven import Context, Request, Response
from httpx import AsyncClient
from orjson import loads

from amebo.decorators.formatters import jsonify
from amebo.decorators.security import authenticated
from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
@authenticated
def tabulate(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    params = ['origin', 'destination', 'gist', 'event', 'completed', 'timeline']
    _origin, _destination, _gist, _event, _completed, _timeline = [req.params.get(p) for p in params]
    _completed = _completed or 'all'

    completed = {'all': '', 'true': 1, 'false': 0}.get(_completed.lower())

    steps = Steps()
    sqls = f'''
        SELECT
            g.rowid as gist, e.microservice as origin, a.event,
            case when
                g.completed <> 0
            then
                'True'
            else
                'False'
            end as
                completed,
            g.timestamped, a.payload, m.microservice as destination
        FROM gists AS g JOIN actions a ON
            g.action = a.action
        JOIN subscribers s ON
            s.subscriber = g.subscriber
        JOIN events e ON
            a.event = e.event
        JOIN microservices m ON
            s.microservice = m.microservice
        {steps.EQUALS(_gist, 'g.rowid')}
        {steps.LIKE(_origin, 'e.microservice')}
        {steps.EQUALS(completed, 'g.completed')}
        {steps.LIKE(_event, 'a.event')}
        {steps.LIKE(_destination, 'm.microservice')}
        {get_timeline(_timeline, steps, column='g.timestamped')}
        ORDER BY g.rowid 
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try:
        cursor = db.cursor()
        rows = cursor.execute(sqls, steps.values).fetchall()
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return
    finally:
        cursor.close()

    res.status = HTTPStatus.OK
    res.body = [{
        'gist': gist,
        'origin': origin,
        'event': event,
        'completed': completed,
        'timestamped': timestamped,
        'payload': loads(payload),
        'destination': destination
    } for gist, origin, event, completed, timestamped, payload, destination in rows]


@jsonify
async def replay(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    id = req.params.get('id')

    try:
        cursor = db.cursor()
        gist = cursor.execute(f'''
            SELECT
                m.location || s.endpoint AS endpoint, a.payload, m.passkey, g.rowid as gid
            FROM gists AS g JOIN actions a ON
                g.action = a.action
            JOIN subscribers s ON
                s.subscriber = g.subscriber
            JOIN events e ON
                a.event = e.event
            JOIN microservices m ON
                s.microservice = m.microservice
            WHERE g.rowid = ?;
        ''', (id,)).fetchone()
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return
    finally:
        cursor.close()
    
    if not gist:
        res.status = HTTPStatus.NOT_FOUND
        res.body = {}
        return

    try:
        sender = AsyncClient()

        endpoint, payload, passkey, gid = gist
        headers = {'Content-Type': 'application/json', 'X-Secret-Key': passkey}

        response = await sender.post(endpoint, json=loads(payload), headers=headers)
        if response.status_code not in [HTTPStatus.ACCEPTED, HTTPStatus.OK]:
            raise ConnectionRefusedError('Endpoint maybe offline, failed to handle gist')
    except ConnectionRefusedError as exc:
        res.status = HTTPStatus.SERVICE_UNAVAILABLE
        try: proxied = response.json()
        except: proxied = None
        res.body = {'gist': gid, 'proxied': proxied, 'error': f'{exc}'}
        return
    except Exception as exc:
        res.status = HTTPStatus.BAD_GATEWAY
        res.body = {'error': f'{exc}'}
        return
    finally:
        await sender.aclose()

    res.satus = HTTPStatus.ACCEPTED
    try: proxied = response.json()
    except: proxied = None
    res.body = {'gist': gid, 'proxied': proxied}
    return



from sqlite3 import Connection, Cursor
from typing import List, TypeVar

from pydantic import BaseModel


T = TypeVar("T")


class Entity(object):
    def __init__(self, model: BaseModel[T], db: Connection):
        self.db = db
        self.table = model.__class__.__name__.lower()
        self.cursor = db.cursor()

    def select_from(self, *fields: List[str]) -> T:
        return self.cursor.execute(f'SELECT {",".join(fields)} from {self.table}').fetchall()

    def select_by_id(self, key, value, *fields, offset=0, limit=25):
        return self.cursor.execute(f"""
            SELECT {','.join(fields)} FROM {self.table}
                WHERE {key} = ?
            OFFSET {offset} LIMIT {limit}
        """, value).fetchall()



from http import HTTPStatus
from sqlite3 import Connection, Cursor

from heaven import Context, Request, Response

from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import expects
from amebo.models.subscribers import Subscriber
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
def tabulate(req: Request, res, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    params = ['id', 'microservice', 'matchrule', 'event', 'endpoint', 'description', 'timeline']
    [
        _id, _microservice, _matchrule,
        _event, _endpoint, _description, _timeline] = [req.params.get(p) for p in params]
    
    m = 'microservice'
    steps = Steps()
    sqls = f'''
        SELECT
            subscriber, event, microservice, endpoint, description, timestamped
        FROM subscribers
            {steps.EQUALS(_id, 'subscriber')}
            {steps.LIKE(_microservice, m)}
            {steps.LIKE(_event, 'event')}
            {steps.LIKE(_endpoint, 'endpoint')}
            {steps.LIKE(_description, 'description')}
            {get_timeline(_timeline, steps)}
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try:
        cursor = db.cursor()
        rows = cursor.execute(sqls, steps.values).fetchall()
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = None
        return
    finally:
        cursor.close()

    res.status = HTTPStatus.OK
    res.body = [{
        'subscriber': subscriber,
        'event': event,
        'microservice': microservice,
        'endpoint': endpoint,
        'description': description,
        'timestamped': timestamped
    } for subscriber, event, microservice, endpoint, description, timestamped in rows]


@jsonify
@expects(Subscriber)
def insert(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    subscriber: Actions = ctx.subscriber

    table = 'subscribers'
    fields = ('event', 'microservice', 'endpoint', 'description', 'timestamped',)
    values = (subscriber.action, subscriber.microservice, subscriber.endpoint, subscriber.description, subscriber.timestamped)

    try:
        cursor: Cursor = db.cursor()
        subscriberid = cursor.execute(f'''
            INSERT INTO {table}({', '.join(fields)}) VALUES(?, ?, ?, ?, ?) RETURNING rowid;
        ''', values).fetchone()
        db.commit()
    except Exception as exc:
        res.status = HTTPStatus.UPGRADE_REQUIRED
        res.body = {'error': f'{exc}'}
        return
    finally:
        cursor.close()

    res.status = HTTPStatus.CREATED
    subscriber.subscriber = subscriberid[0]
    res.body = subscriber.dict()



from datetime import datetime
from json import loads
from typing import Optional, Union

from pydantic import AnyHttpUrl
from pydantic import BaseModel as Model
from pydantic import Field, field_validator


class Action(Model):
    action: str
    microservice: str
    schemata: Union[dict, str]
    timestamped: datetime = Field(default_factory=datetime.now)

    @classmethod
    @field_validator('action')
    def validate_action(cls, value: str):
        if len(value) < 3: raise ValueError()
        return value

    @classmethod
    @field_validator('schemata')
    def validate_schemata(cls, value: Union[str, dict]):
        if isinstance(value, dict): return value
        return loads(value)



from datetime import datetime
from json import loads
from typing import Optional, Union

from pydantic import AnyHttpUrl
from pydantic import BaseModel as Model
from pydantic import Field, field_validator


class Credential(Model):
    username: str
    password: str
    scheme: str


class Microservice(Model):
    microservice: str
    address: AnyHttpUrl
    passkey: str = Field(min_length=32)
    timestamped: datetime = Field(default_factory=datetime.now)

    @classmethod
    @field_validator('microservice')
    def no_spaces(cls, val: str) -> str:
        if val.count(' '): raise ValueError('microservice name can not contain spaces')
        return val

    @property
    def password(self):
        return '***'


class Location(Model):
    location: AnyHttpUrl
    passkey: str



from os import environ
from uuid import uuid4

# installed libs
from dotenv import load_dotenv
from heaven import Application
from heaven.constants import STARTUP, SHUTDOWN

# src code
from amebo.aproko import aproko
from amebo.constants.literals import SECRET_KEY


router = Application({
    'env_loaded': load_dotenv(),
    'db': environ.get('WASHIKA_STORE') or 'sqlite',
    'envelope_size': int(environ.get('ENVELOPE_SIZE') or 256),  # how many tasks to fetch at once for processing
    'idles': 5,  # sleep for 5 seconds
    'rest_when': 0,  # reduce frequency of daemons when tasks less than 5
    SECRET_KEY: environ.get('secret_key') or str(uuid4())
})


# setup daemons and app engine helpers
router.daemons = aproko
router.ASSETS('public', relative_to=__file__)
router.TEMPLATES('templates', relative_to=__file__)


# set up hooks
router.ON(STARTUP, 'amebo.middlewares.database.connect')
router.ON(STARTUP, 'amebo.middlewares.database.cache')
router.ON(SHUTDOWN, 'amebo.middlewares.database.disconnect')
router.ON(STARTUP, 'amebo.middlewares.database.initialize')
router.ON(STARTUP, 'amebo.middlewares.security.upsudo')


# hooks
router.BEFORE('/*', 'amebo.middlewares.security.cors')


# web ui- views/pages/screens
router.GET('/', 'amebo.controllers.app.login')
router.GET('/p/:page', 'amebo.controllers.app.pages')


# api
router.GET('/v1/actions', 'amebo.controllers.actions.tabulate')
router.GET('/v1/events', 'amebo.controllers.events.tabulate')
router.GET('/v1/microservices', 'amebo.controllers.microservices.tabulate')
router.GET('/v1/subscribers', 'amebo.controllers.subscribers.tabulate')
router.GET('/v1/gists', 'amebo.controllers.gists.tabulate')
router.POST('/v1/tokens', 'amebo.controllers.microservices.authenticate')
router.POST('/v1/actions', 'amebo.controllers.actions.insert')
router.POST('/v1/events', 'amebo.controllers.events.insert')
router.POST('/v1/microservices', 'amebo.controllers.microservices.insert')
router.POST('/v1/subscribers', 'amebo.controllers.subscribers.insert')
router.POST('/v1/gists/:id', 'amebo.controllers.gists.replay')
router.PUT('/v1/microservices/:id', 'amebo.controllers.microservices.update')

# maybe add a route to clear cache of compiled schemas ?


# comment me out in production
# router.POST('/h1/identity-created', amebo_sleeper)



"""
TODO: change to supersql and group sqlite&pg -> sql so the final option is sql | redis
"""

from constants.literals import PG, REDIS, SQLITE


def list_actions_redis():
    pass


def list_actions_sql():
    pass


def list_applications_redis():...


def list_applications_sql():...


def list_events_redis():
    pass


def list_events_sql():
    pass


def list_subscribers_redis():...


def list_subscribers_sql():...


list_actions = {
    REDIS: list_actions_redis,
    SQLITE: list_actions_sql
}


list_events = {
    REDIS: list_events_redis,
    SQLITE: list_events_sql
}


list_applications = {
    REDIS: list_applications_redis,
    SQLITE: list_applications_sql
}


list_subscribers = {
    REDIS: list_subscribers_redis,
    SQLITE: list_subscribers_sql
}
